/*
 * Panasync Tools
 *
 * Copyright (C) 2000 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */

#include <panasync.h>
#include <panadefaults.h>

#include <fstream.h>
#include <getopt.h>	// getopt
#include <stdlib.h>	// random
#include <sys/times.h>	// times
#include <sys/stat.h>	// file stat info
#include <unistd.h>	// access
#include <fcntl.h>	// open, creat
#include <stdio.h>	// sprintf
#include <sys/param.h>	// MAXPATHLEN	(glib.h?)

#include <xtl.h>	// XTL
#include <zlib.h>	// compression lib

extern "C" {
#include <md5.h>	// MD5 digests
}

//	=======================================================================
//	Definition of constants and initialization of global variables
//	=======================================================================
//

const unsigned MD5BUFSIZE	= 16;
const unsigned PTBUFSIZE	= 1024 * 64;

//	=======================================================================
//	Definition of Panasync Tools data structures stored in PDS files
//	=======================================================================
//

//	Definition of filedata structure to be kept in PDS files (XTL enabled)
//
struct	filedata
{
	unsigned char	md5[MD5BUFSIZE];
	pts		stamp;

	template <class Stream> void composite (Stream & stream)
		{ stream.vector(md5, sizeof(md5)).simple(stamp); }

	filedata() {}
	filedata(const unsigned char srcmd5[MD5BUFSIZE], const pts & srcstamp)
		: stamp(srcstamp)
		{ memcpy(md5, srcmd5, sizeof(md5)); }
	filedata(const unsigned char srcmd5[MD5BUFSIZE], string srckind,
		string srcid, string srccounter)
		: stamp(srckind, srcid, srccounter)
		{ memcpy(md5, srcmd5, sizeof(md5)); }
};

//	Definition of mapping type between file names and filedata structures
//	to be kept in PDS files
//
typedef map<string, filedata> panadir_t;

//	=======================================================================
//	Read/Write Panasync's data strucutres from/to PDS files
//	=======================================================================
//

//	Write Panasync's data structures to a PDS file
//
int	write_pdsfile(const char * file, panadir_t & dir)
{
	// Establish externalization environment
	//
	unsigned char	fmb[PTBUFSIZE];		// memory buffer
	mem_buffer	xmb(fmb, sizeof(fmb));	// XTL memory buffer object
	GIOP_format<mem_buffer> gfmb(xmb);
	obj_output<GIOP_format<mem_buffer> > os(gfmb);

	// Establish compression environment
	//
	Bytef		cbuf[PTBUFSIZE];	// Compression buffer
	unsigned long	csize = PTBUFSIZE;	// Compression buffer size

	// Externalize Panasync's data strutures
	//
	os.simple(dir);

	// Compress externalization buffer
	// 
	// 	Compress levels:
	// 		Z_BEST_COMPRESSION,
	// 		Z_BEST_SPEED,
	// 		Z_DEFAULT_COMPRESSION
	//
	assert(compress2(cbuf, &csize, (Bytef *)xmb.data(), 
                  xmb.size(), Z_BEST_COMPRESSION) == Z_OK);

	// Write buffer to file	
	//
	ofstream fp(file);
//	fp.write(xmb.data(), xmb.size());
	fp.write(cbuf, csize);

	return 0;
}

//	Read Panasync's data structures to a PDS file
//
int	read_pdsfile(const char *file, panadir_t & dir)
{
	// Establish externalization environment
	//
	unsigned char	fmb[PTBUFSIZE];		// memory buffer

	// Establish compression environment
	//
	Bytef		cbuf[PTBUFSIZE];	// Compression buffer
	unsigned long	csize = PTBUFSIZE;	// Compression buffer size

	// Read buffer from file
	//
	ifstream fp(file);
//	fp.read(fmb, sizeof(fmb));
	fp.read(cbuf, sizeof(cbuf));

	// Uncompress externatization buffer
	// 
	assert(uncompress((Bytef *)fmb, &csize, cbuf, PTBUFSIZE) == Z_OK);

	// Revert externalization of Panasync's data structures
	//
	mem_buffer	xmb(fmb, sizeof(fmb));	// XTL memory buffer object
	GIOP_format<mem_buffer> gfmb(xmb);
	obj_input<GIOP_format<mem_buffer> > is(gfmb);

	dir.clear();
	is.simple(dir);

	return 0;
}

//	========================================================================
//	File name manipulation and inode reporting auxiliary functions
//	========================================================================
//

//	Returns file name component of 'path'
//
char *	getbasename(const char * path, char * _name)
{
	char * p = strrchr(path, '/');
	strcpy(_name, p == NULL ? path : p + 1);
	return _name;
}

//	Returns directory path component of 'path'
//	(or './' if directory path not specified)
//
char *	getbasedir(const char * path, char * _path)
{
	char * p = strrchr(path, '/');
	if (p != NULL)
	{
		strncpy(_path, path, p - path + 1);
		_path[p - path + 1]=0;
	}
	else
		strcpy(_path, "./");
	return _path;
}

//	Returns the i-node of 'path'
//
ino_t	getinode(const char * path)
{
	struct stat fstat;
	stat(path, &fstat);
	return fstat.st_ino;
}

//	Returns true if 'path' is a directory
//
bool	isdir(const char * path)
{
	struct stat fstat;
	stat(path, &fstat);
	return S_ISDIR(fstat.st_mode);
}

//	========================================================================
//	panainit	- initialize Panasync Tools
//	========================================================================
//

void	display_panainit_usage()
{
	cerr <<
"usage: panainit {--help -h}\n"
"       panainit {--unique_base -u} url [--base_file -b] [file]\n"
"       panainit {--string_base -s} phrase [--base_file -b] [file]\n"
"       panainit {--random_base -r} [--base_file -b] [file]\n";
}

void	panainit(int argc, char ** argv)
{
	// By default panainit should generate a unique-based timestamp
	// but for now it's convenient to default to random.
	string	pt_base_kind	= "random";
	string	pt_base_id	= "";
	char * 	pt_base_path	= PTDEF_BASE_PATH;

	// Process command-line options
	//
	while (1)
	{
		int short_option;	// Short option character

		// Long options declaration & short options mappings
		//
		static struct option long_options[] =
		{
			{ "unique_base", 1, NULL, 'u' },
			{ "string_base", 1, NULL, 's' },
			{ "random_base", 0, NULL, 'r' },
			{ "base_file", 1, NULL, 'b' },
			{ "help", 0, NULL, 'h' },
			{ 0, 0, 0, 0 }
		};

		// Check end of option list
		//
		if ((short_option = getopt_long(argc, argv, "u:s:rb:h", 
			long_options, NULL)) == -1) break;

		// Process short option
		//
		switch (short_option)
		{
			case 'u':
				pt_base_kind = "unique";
				//pt_base_id = get_base_id();
				cerr << "panainit: 'unique_base' not ";
				cerr << "implemented ";
				cerr << "[internal error]\n";
				exit(2);
				break;
			case 's':
				pt_base_id = optarg;
				pt_base_kind = "string";
				break;
			case 'r':
				char tmp[32 + 1];
				srandom(times(NULL));
				for (int i = 0; i != 32; i++)
					tmp[i] = random() % 24 + 'a';
				tmp[32] = '\0';
				pt_base_id = tmp;
				pt_base_kind = "random";
				break;
			case 'b':
				pt_base_path = strdup(optarg);
				break;
			case 'h':
				display_panainit_usage();
				exit(0);
				break;
			case ':':
				cerr << "panainit: missing parameter on '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			case '?':
				cerr << "panainit: extraneous parameter '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			default:
				cerr << "panainit: unexpected option '";
				cerr << short_option << "' ";
				cerr << "[internal error]\n";
				exit(2);
		}
	}

	// Check unimplemented options
	//
	if (pt_base_kind == "unique")
	{
		cerr << "panainit: '--unique_base' ";
		cerr << "not implemented [internal error]\n";
		exit(2);
	}

   // Ensure read/write access to the target directory that stores base
	//
	char basedir[MAXPATHLEN];
	getbasedir(pt_base_path, basedir);
	if (access(basedir, R_OK+W_OK) == -1)
	{
		cerr << "panainit: " << basedir << " cannot be accessed. Create a writable " << basedir << " directory or use option -b to specify a diferent directory\n";
		exit(1);
	}

	// Create new Panasync base object
	//
	// The base counter initialization is slightly different
	// from within panainit and pananew -- we may want to be
	// able to distinguish random-based generated timestamps
	// in the future. The base counter is initialized with
   	// "1" in panainit and set to "0" in pananew.

	pts::base	newbase = { pt_base_kind, pt_base_id, "1" };

	// Establish externalization environment
	//
	char		buf[PTBUFSIZE];
	mem_buffer	mb(buf, sizeof(buf));
	GIOP_format<mem_buffer> gfmb(mb);	
	obj_output<GIOP_format<mem_buffer> > os(gfmb);

	// Externalize Panasync's base
	//
	os.simple(newbase.kind).simple(newbase.id).simple(newbase.counter);

	// Write externalized Panasync base
	//
	ofstream	basefile(pt_base_path);
	basefile.write(mb.data(), mb.size());

	// Free resources
	//
	if (pt_base_path != PTDEF_BASE_PATH) free(pt_base_path);
}

//	========================================================================
//	pananew
//	========================================================================
//

void	display_pananew_usage()
{
	cerr <<
"usage: pananew {--help -h}\n"
"       pananew [--base_file -b] [file] [--force -f] [source] target\n";
}

void	pananew(int argc, char ** argv)
{
	char *	pt_source	= 0;
	char *	pt_target	= 0;
	bool	pt_force	= false;
	char * 	pt_base_path	= PTDEF_BASE_PATH;

	// Process command-line options
	//
	while (1)
	{
		int short_option;	// Short option character

		// Long options declaration & short options mappings
		//
		static struct option long_options[] =
		{
			{ "base_file", 1, NULL, 'b' },
			{ "force", 0, NULL, 'f' },
			{ "help", 0, NULL, 'h' },
			{ 0, 0, 0, 0 }
		};

		// Check end of option list
		//
		if ((short_option = getopt_long (argc, argv, "b:fh", 
			long_options, NULL)) == -1) break;

		// Process short option
		//
		switch (short_option)
		{
			case 'b':
				pt_base_path = strdup(optarg);
				break;
			case 'f':
				pt_force = true;
				break;
			case 'h':
				display_pananew_usage();
				exit(0);
				break;
			case ':':
				cerr << "pananew: missing parameter on '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			case '?':
				cerr << "pananew: extraneous parameter '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			default:
				cerr << "pananew: unexpected option '";
				cerr << short_option << "' ";
				cerr << "[internal error]\n";
				exit(2);
		}
	}

	// Retrieve target (and optional source) file from command-line
	//
	if (optind + 2 == argc)
	{
		pt_source = argv[optind++];
		pt_target = argv[optind];
	}
	else
	if (optind + 1 == argc)
	{
		pt_target = argv[optind];
	}
	else
	{
		cerr << "pananew: illegal command-line arguments.\n";
		exit(1);
	}

	// Ensure read access to the optional source file
	//
	if (pt_source && access(pt_source, R_OK) == -1)
	{
		cerr << "pananew: cannot read from '" << pt_source << "'\n";
		exit(1);
	}

	// Ensure read/write access to the target directory
	//
	char basedir[MAXPATHLEN];
	getbasedir(pt_target, basedir);
	if (access(basedir, R_OK+W_OK) == -1)
	{
		cerr << "pananew: " << basedir << " cannot be accessed.\n";
		exit(1);
	}

	// Ensure existance and read/write access to the target pdsfile
	//
	char pdsfile[MAXPATHLEN];
	sprintf(pdsfile, "%s%s", basedir, PTDEF_PDS_FILENAME);
	if (access(pdsfile, F_OK) == 0 && access(pdsfile, R_OK+W_OK) == -1)
	{
		cerr << "pananew: " << pdsfile << " cannot be ";
		cerr << "accessed for read/write.\n";
		exit(1);
	}

	//streamlock slpdsfile(pdsfile);

//	cout << "ANTES DO READ_PDSFILE (" << pdsfile << ")\n";

	// Read target pdsfile if present
	//
	panadir_t dir;
	if (access(pdsfile, F_OK) == 0)
		read_pdsfile(pdsfile, dir);

//	cout << "DEPOIS DO READ_PDSFILE\n";

	// Check if target already exists
	//
	char basename[MAXPATHLEN];
       	getbasename(pt_target, basename);
	if (dir.find(basename) != dir.end() && !pt_force)
	{
		cerr << "pananew: target " << pt_target << " exists. ";
		cerr << "Use --force.\n";
		exit(1);
	}

	//streamlock slbasefile(base_file);

	// Panasync base object
	//
	pts::base base = { "random", "", "0" };

	if (access(pt_base_path, F_OK) == -1)
	{
		char tmp[32 + 1];
		srandom(times(NULL));
		for (int i = 0; i != 32; i++)
			tmp[i] = random() % 24 + 'a';
		tmp[32] = '\0';
		base.id = tmp;
	}
	else
	{
		// Read Panasync's base file
		//
		if (access(pt_base_path, R_OK) == -1)
		{
			cerr << "pananew: cannot read from " << pt_base_path;
			cerr << "\n";
			exit(1);
		}

		char buf[PTBUFSIZE];		// Change into a XTL autobuffer!
		ifstream basefilein(pt_base_path);
		basefilein.read(buf, sizeof(buf));
		basefilein.close();

		// Reverse externalization
		//
		mem_buffer mb(buf, sizeof(buf));
		GIOP_format<mem_buffer> gfmb(mb);	
		obj_input<GIOP_format<mem_buffer> > is(gfmb);
		is.simple(base.kind).simple(base.id).simple(base.counter);

		// Increment base counter
		int tmp = atoi(base.counter.c_str());
		tmp++;
		char straux[PTBUFSIZE];
		sprintf(straux, "%d", tmp);
		base.counter = string(straux);

		// Write Panasync's base file
		//
		mb.rewind();
		ofstream basefileout(pt_base_path);
		obj_output<GIOP_format<mem_buffer> > os(gfmb);
		os.simple(base.kind).simple(base.id).simple(base.counter);
		basefileout.write(mb.data(), mb.size());
		basefileout.close();
	}

	if (pt_source == 0)
	{
		// Create target file
		//
		ofstream tmp(pt_target);
	}
	else
	{
		// Copy source content to target file
		//
		if (access(pt_target, F_OK) == -1
			|| getinode(pt_source) != getinode(pt_target))
		{
			int ifs = open(pt_source, O_RDONLY);
			int ofs = creat(pt_target, 0666), n;
			char buffer[1024];
			while ((n = read(ifs, buffer, sizeof(buffer))) > 0)
				assert(write (ofs, buffer, n) == n);
			close (ifs);
			close (ofs);
		}
	}

	// Calculate md5 digest for target file
	//
	FILE * fp = fopen(pt_target, "r");
	unsigned char md5[16];
	md5_stream(fp, md5);
	fclose(fp);
	dir[basename] = filedata(md5, base.kind, base.id, base.counter);

	// Update target pdsfile
	//
	write_pdsfile(pdsfile, dir);
	//	panadir_t dir2;
	//	read_pdsfile(pdsfile, dir2);
	//	for (panadir_t::iterator i = dir2.begin(); i != dir2.end(); i++)
	//	{
	//		cout << "pananew: " << i->first.c_str() << endl;
	//	}
	//	pds();

	// Free resources
	//
	if (pt_base_path != PTDEF_BASE_PATH) free(pt_base_path);
}

//	========================================================================
//	panadup
//	========================================================================
//

void	display_panadup_usage()
{
	cerr <<
"usage: panadup {--help -h}\n"
"       panadup [--base_file -b] [file] [--force -f] source target\n";
}

void	panadup(int argc, char ** argv)
{
	char *	pt_source	= 0;
	char *	pt_target	= 0;
	bool	pt_force	= false;
	char * 	pt_base_path	= PTDEF_BASE_PATH;

	// Process command-line options
	//
	while (1)
	{
		int short_option;	// Short option character

		// Long options declaration & short options mappings
		//
		static struct option long_options[] =
		{
			{ "base_file", 1, NULL, 'b' },
			{ "force", 0, NULL, 'f' },
			{ "help", 0, NULL, 'h' },
			{ 0, 0, 0, 0 }
		};

		// Check end of option list
		//
		if ((short_option = getopt_long (argc, argv, "b:fh", 
			long_options, NULL)) == -1) break;

		// Process short option
		//
		switch (short_option)
		{
			case 'b':
				pt_base_path = strdup(optarg);
				break;
			case 'f':
				pt_force = true;
				break;
			case 'h':
				display_panadup_usage();
				exit(0);
				break;
			case ':':
				cerr << "panadup: missing parameter on '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			case '?':
				cerr << "panadup: extraneous parameter '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			default:
				cerr << "panadup: unexpected option '";
				cerr << short_option << "' ";
				cerr << "[internal error]\n";
				exit(2);
		}
	}

	// Retrieve target (and optional source) file from command-line
	//
	if (optind + 2 == argc)
	{
		pt_source = argv[optind++];
		pt_target = argv[optind];
	}
	else
	{
		cerr << "panadup: illegal command-line arguments.\n";
		exit(1);
	}

	char srcbasename[MAXPATHLEN];
	char srcbasedir[MAXPATHLEN];
	char trgbasename[MAXPATHLEN];
	char trgbasedir[MAXPATHLEN];
	getbasename(pt_source, srcbasename);
	getbasedir(pt_source, srcbasedir);
	getbasename(pt_target, trgbasename);
	getbasedir(pt_target, trgbasedir);

	// Ensure read access to the source file
	//
	if (access(pt_source, R_OK) == -1)
	{
		cerr << "panadup: cannot read from " << pt_source << "\n";
		exit(1);
	}

	// Ensure read access to the source directory
	//
	if (access(srcbasedir, R_OK) == -1)
	{
		cerr << "panadup: " << srcbasedir << " cannot be read.\n";
		exit(1);
	}

	// Ensure read/write access to the source pdsfile
	//
	char srcpdsfile[MAXPATHLEN];
        strcpy(srcpdsfile, srcbasedir);
        strcat(srcpdsfile, PTDEF_PDS_FILENAME);
	if (access(srcpdsfile, R_OK+W_OK) == -1)
	{
		cerr << "panadup: " << srcpdsfile;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	// Read source pdsfile
	//
	panadir_t srcdir;
	read_pdsfile(srcpdsfile, srcdir);

	// Ensure source is a panasync file
	//
	if (srcdir.find(srcbasename) == srcdir.end())
	{
		cerr << "panadup: source " << pt_source;
		cerr << " is not a panasync file.\n";
		exit(1);
	}

	// Calculate md5 digest of source file
	//
	unsigned char md5[16];
	FILE* fp=fopen(pt_source, "r");
	md5_stream(fp, md5);
	fclose(fp);

	// Check content modification of the source file
	// and set it dirty
	//
	if (memcmp(md5, srcdir[srcbasename].md5, sizeof(md5)) != 0)
	{
		filedata srcfdata(md5, srcdir[srcbasename].stamp);
		srcdir.erase(srcbasename);
		srcfdata.stamp.set_dirty();
		srcdir[srcbasename]=srcfdata;
	}

	// Ensure read/write access to the target directory
	//
	if (access(trgbasedir, R_OK+W_OK) == -1)
	{
		cerr << "panadup: " << trgbasedir;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	// Check existance of target pdsfile
	// if it exists ensure read/write access
	//
	char trgpdsfile[MAXPATHLEN];
        strcpy(trgpdsfile, trgbasedir);
        strcat(trgpdsfile, PTDEF_PDS_FILENAME);
	if (access(trgpdsfile, F_OK) == 0
		&& access(trgpdsfile, R_OK+W_OK) == -1)
	{
		cerr << "panadup: " << trgpdsfile;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	//streamlock slpdsfile(pdsfile);

	// Read pdsfile if exists
	//
	panadir_t trgdir;
	if (access(trgpdsfile, F_OK) == 0)
		read_pdsfile(trgpdsfile, trgdir);

	// Check existance of target file
	//
	if (trgdir.find(trgbasename) != trgdir.end())
	{
		if (pt_force)
		{
			// Delete target file
			//
			trgdir.erase(trgbasename);
			if (access(pt_target, F_OK) == 0)
				unlink(pt_target);
		}
		else
		{
			cerr << "panadup: target " << pt_target << " exists.\n";
			cerr << "You can use --force, but with caution, since";
			cerr << " a leak on time stamp space will occur.\n";
			cerr << "Shouldnt you panajoin " << pt_target;
			cerr << " or use a diferent target name?\n";
			exit(1);
		}
	}

	// Fork source file
	//
	filedata srcfdata = srcdir[srcbasename];
	filedata trgfdata(srcfdata.md5, srcfdata.stamp.fork());

	// Ensure source and target are different files
	//
	if (access(pt_target, F_OK) == 0
		&& getinode(pt_source) == getinode(pt_target))
		{
			cerr << "panadup: source " << pt_source;
			cerr << " and target " << pt_target;
			cerr << " are the same file.\n";
			exit(1);
		}
	
	// Update pdsfiles, but first check if pdsfiles are shared
	//
	if (access(trgpdsfile, F_OK) == 0
		&& getinode(srcpdsfile) == getinode(trgpdsfile))
	{
		if (strcmp(srcbasename, trgbasename) == 0)
		{
			cerr << "panadup: source " << pt_source;
			cerr << " and target " << pt_target;
			cerr << " would be the same file.\n";
			exit(1);
		}
		srcdir[srcbasename] = srcfdata;
		srcdir[trgbasename] = trgfdata;
		write_pdsfile(srcpdsfile, srcdir);
	}
	else
	{
		srcdir[srcbasename] = srcfdata;
		trgdir[trgbasename] = trgfdata;
		write_pdsfile(srcpdsfile, srcdir);
		write_pdsfile(trgpdsfile, trgdir);
	}

	// Copy source content to target file
	//
	int ifs = open(pt_source, O_RDONLY);
	int ofs = creat(pt_target, 0666), n;
	char buffer[1024];
	while ((n = read(ifs, buffer, sizeof(buffer))) > 0)
		assert(write (ofs, buffer, n) == n);
	close (ifs);
	close (ofs);

	// Free resources
	//
	if (pt_base_path != PTDEF_BASE_PATH) free(pt_base_path);
}

//	========================================================================
//	panajoin
//	========================================================================
//

void	display_panajoin_usage()
{
	cerr <<
"usage: panajoin {--help -h}\n"
"       panajoin [--base_file -b] [file] [--content -c] [contentfile]\n"
"                   [--older -o] source sourcetarget\n";
}

void	panajoin(int argc, char ** argv)
{
	char *	pt_source	= 0;
	char *	pt_target	= 0;
	char *	pt_content	= 0;
	bool	pt_older	= false;
	char * 	pt_base_path	= PTDEF_BASE_PATH;

	// Process command-line options
	//
	while (1)
	{
		int short_option;	// Short option character

		// Long options declaration & short options mappings
		//
		static struct option long_options[] =
		{
			{ "base_file", 1, NULL, 'b' },
			{ "content", 1, NULL, 'c' },
			{ "older", 0, NULL, 'o' },
			{ "help", 0, NULL, 'h' },
			{ 0, 0, 0, 0 }
		};

		// Check end of option list
		//
		if ((short_option = getopt_long (argc, argv, "b:c:oh", 
			long_options, NULL)) == -1) break;

		// Process short option
		//
		switch (short_option)
		{
			case 'b':
				pt_base_path = strdup(optarg);
				break;
			case 'c':
				strcpy(pt_content, optarg);
				break;
			case 'o':
				pt_older = true;
				break;
			case 'h':
				display_panajoin_usage();
				exit(0);
				break;
			case ':':
				cerr << "panajoin: missing parameter on '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			case '?':
				cerr << "panajoin: extraneous parameter '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			default:
				cerr << "panajoin: unexpected option '";
				cerr << short_option << "' ";
				cerr << "[internal error]\n";
				exit(2);
		}
	}

	// Retrieve target (and optional source) file from command-line
	//
	if (optind + 2 == argc)
	{
		pt_source = argv[optind++];
		pt_target = argv[optind];
	}
	else
	{
		cerr << "panajoin: illegal command-line arguments.\n";
		exit(1);
	}

	char srcbasename[MAXPATHLEN];
	char srcbasedir[MAXPATHLEN];
	char trgbasename[MAXPATHLEN];
	char trgbasedir[MAXPATHLEN];
	getbasename(pt_source, srcbasename);
	getbasedir(pt_source, srcbasedir);
	getbasename(pt_target, trgbasename);
	getbasedir(pt_target, trgbasedir);

	// Ensure read access to the source file
	//
	if (access(pt_source, R_OK) == -1)
	{
		cerr << "panajoin: cannot read from " << pt_source << "\n";
		exit(1);
	}

	// Ensure read/write access to the source directory
	//
	if (access(srcbasedir, R_OK+W_OK) == -1)
	{ cerr << srcbasedir << " cannot be accessed.\n"; exit(1); }

	// Ensure read/write access to the source pdsfile
	//
	char srcpdsfile[MAXPATHLEN];
        strcpy(srcpdsfile, srcbasedir);
        strcat(srcpdsfile, PTDEF_PDS_FILENAME);
	if (access(srcpdsfile, R_OK+W_OK) == -1)
	{
		cerr << "panajoin: " << srcpdsfile;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	// Read source pdsfile
	//
	panadir_t srcdir;
	read_pdsfile(srcpdsfile, srcdir);

	// Ensure source is a panasync file
	//
	if (srcdir.find(srcbasename) == srcdir.end())
	{
		cerr << "panajoin: source " << pt_source;
		cerr << " is not a panasync file.\n";
		exit(1);
	}

	// Calculate md5 digest of source file
	//
	unsigned char srcmd5[16];
	FILE* srcfp=fopen(pt_source, "r");
	md5_stream(srcfp, srcmd5);
	fclose(srcfp);

	// Check content modification of the source file
	// and set it dirty
	//
	filedata srcfdata(srcmd5, srcdir[srcbasename].stamp);
	if (memcmp(srcmd5, srcdir[srcbasename].md5, sizeof(srcmd5)) != 0)
	{
		srcdir.erase(srcbasename);
		srcfdata.stamp.set_dirty();
		srcdir[srcbasename]=srcfdata;
	}

	// Ensure read access to the target file
	//
	if (access(pt_target, R_OK) == -1)
	{
		cerr << "panajoin: cannot read from " << pt_target << "\n";
		exit(1);
	}

	// Ensure read/write access to the target directory
	//
	if (access(trgbasedir, R_OK+W_OK) == -1)
	{
		cerr << "panajoin: " << trgbasedir;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	// Ensure source and target are different files
	//
	if (getinode(pt_source) == getinode(pt_target))
	{
		cerr << "panajoin: source " << pt_source;
		cerr << " and sourcetarget " << pt_target;
		cerr << " are the same file.\n";
		exit(1);
	}

	// Check existance of target pdsfile
	// if it exists ensure read/write access
	//
	char trgpdsfile[MAXPATHLEN];
        strcpy(trgpdsfile, trgbasedir);
        strcat(trgpdsfile, PTDEF_PDS_FILENAME);
	if (access(trgpdsfile, R_OK+W_OK) == -1)
	{
		cerr << "panajoin: " << trgpdsfile;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	// Read target pdsfile
	//
	panadir_t trgdir;
	read_pdsfile(trgpdsfile, trgdir);

	// Calculate md5 digest of target file
	//
	unsigned char trgmd5[16];
	FILE* trgfp=fopen(pt_target, "r");
	md5_stream(trgfp, trgmd5);
	fclose(trgfp);

	// Check source and target pdsfile are the same
	//
	bool samepdsfile = getinode(srcpdsfile) == getinode(trgpdsfile);
	
	// Ensure different file names
	//
       	if (samepdsfile && strcmp(srcbasename, trgbasename) == 0)
	{
		cerr << "panajoin: source " << pt_source;
		cerr << " and sourcetarget " << pt_target;
		cerr << " are the same file.\n";
		exit(1);
	}

	// Check content modification of the target file
	// and set it dirty
	//
	filedata trgfdata(trgmd5, trgdir[trgbasename].stamp);
	if (memcmp(trgmd5, trgdir[trgbasename].md5, sizeof(trgmd5)) != 0)
	{
		trgdir.erase(trgbasename);
		trgfdata.stamp.set_dirty();
		if (!samepdsfile)
			trgdir[trgbasename]=trgfdata;
		else
		srcdir[trgbasename]=trgfdata;
	}

	//streamlock slpdsfile(pdsfile);

	if (pt_content == 0)
	{
		if (srcfdata.stamp.equivalent(trgfdata.stamp))
		{
			// Equivalent
			//
			pt_content = strdup(pt_target);
		}
		else
		{
			// Check domination between source and target files
			//
			if (srcfdata.stamp.dominates(trgfdata.stamp))
			{
				// Source is newer
				//
				pt_content = pt_older ? strdup(pt_target) 
					: strdup(pt_source);
			}
			else
			{
				if (trgfdata.stamp.dominates(srcfdata.stamp))
				{
					// Target is newer
					//
					pt_content =
						pt_older ? strdup(pt_source)
						: strdup(pt_target);
				}
				else
				{
					// Concurrency
					//
					cerr << "panajoin: files are ";
					cerr << "concurrent.\n";
					cerr << "Use --content to specify ";
					cerr << "substitute content\n";
					exit(1);
				}
			}
		}
	}

	// Copy content
	//
	if (getinode(pt_content) != getinode(pt_target))
	{
		int ifs = open(pt_content, O_RDONLY);
		int ofs = creat(pt_target, 0666), n;
		char buffer[1024];
		while ((n = read(ifs, buffer, sizeof(buffer))) > 0)
			assert(write (ofs, buffer, n) == n);
		close (ifs);
		close (ofs);
	}

	// Calculate md5 digest of target file
	//
	if (getinode(pt_content) != getinode(pt_target))
	{
		trgfp=fopen(pt_target, "r");
		md5_stream(trgfp, trgmd5);
		fclose(trgfp);
	}

	// Join timestamps
	//
	filedata newfdata(trgmd5, srcfdata.stamp.join(trgfdata.stamp));
	srcdir.erase(srcbasename);
	if (!samepdsfile)
	{
		trgdir.erase(trgbasename);
		trgdir[trgbasename] = newfdata;
		write_pdsfile(trgpdsfile, trgdir);
	}
	else
	{
		srcdir.erase(trgbasename);
		srcdir[trgbasename] = newfdata;
	}
	write_pdsfile(srcpdsfile, srcdir);

	unlink(pt_source);
}


//	========================================================================
//	panalist
//	========================================================================
//

void	display_panalist_usage()
{
	cerr << 
"usage: panalist {--help -h}\n"
"       panalist [--compare_file -c] [file] [--long -l] [directory]\n";
}

void	panalist(int argc, char ** argv)
{
	char pdsfile[MAXPATHLEN];
	char eachfile[MAXPATHLEN];
	char tmpfilename[MAXPATHLEN];
	bool longopt = false;
	bool isfile;
	char * compare_file = 0;

	// Process command-line options
	//
	while (1)
	{
		int short_option;	// Short option character

		// Long options declaration & short options mappings
		//
		static struct option long_options[] =
		{
			{ "compare", 1, NULL, 'c' },
			{ "help", 0, NULL, 'h' },
			{ "long", 0, NULL, 'l' },
			{ 0, 0, 0, 0 }
		};

		// Check end of option list
		//
		if ((short_option = getopt_long (argc, argv, "c:hl", 
			long_options, NULL)) == -1) break;

		// Process short option
		//
		switch (short_option)
		{
			case 'c':
				compare_file = strdup(optarg);
				break;
			case 'l':
				longopt = true;
				break;
			case 'h':
				display_panalist_usage();
				exit(0);
				break;
			case ':':
				cerr << "panalist: missing parameter on '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			case '?':
				cerr << "panalist: extraneous parameter '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			default:
				cerr << "panalist: unexpected option '";
				cerr << short_option << "' ";
				cerr << "[internal error]\n";
				exit(2);
		}
	}

	if (compare_file) cout << "Comparing " << compare_file << endl;
	// get target pdsfiles
	//
	do
	{
		//// cout << optind << " " << argc << endl;
		if (optind == argc)
			sprintf(pdsfile, "./%s", PTDEF_PDS_FILENAME);
		else 
		{
			// cout << "(" << argv[optind] << ")\n";
			// Check is it is a file or a directory
			//
			if (isdir(argv[optind]))
			{
				char * slashpointer=argv[optind]+strlen(argv[optind])-1;
				if( *slashpointer == '/' ) *slashpointer='\0'; 
				sprintf(pdsfile, "%s/%s", 
					argv[optind], PTDEF_PDS_FILENAME);
				isfile=false;
			}
			else
			{
				char tmp[MAXPATHLEN];
				getbasedir(argv[optind], tmp);
				char * slashpointer=tmp+strlen(tmp)-1;
				if( *slashpointer == '/' ) *slashpointer='\0'; 
				sprintf(pdsfile, "%s/%s", 
					tmp, PTDEF_PDS_FILENAME);
				getbasename(argv[optind],eachfile);
				isfile=true;
			}
		}

		// check for compare_file
		//
		panadir_t cmpdir;
		char cmpbasename[MAXPATHLEN];
		char cmpbasedir[MAXPATHLEN];

		if ( compare_file != 0 )
		{
			// ensure read access to optional compare_file
			// (for future md5 check)
			//
			if (access(compare_file, R_OK) == -1)
			{ cerr << "cannot read from " << compare_file << "\n"; exit(1); }

			// ensure read access to the compare_file directory
			//
			getbasedir(compare_file, cmpbasedir);
			if (access(cmpbasedir, R_OK) == -1)
			{ cerr << cmpbasedir << " cannot be accessed\n"; exit(1); }

			// ensure read access to the compare_file pdsfile
			//
			char cmppdsfile[MAXPATHLEN];
				  strcpy(cmppdsfile, cmpbasedir);
				  strcat(cmppdsfile, PTDEF_PDS_FILENAME);
			if (access(cmppdsfile, R_OK) == -1)
			{ cerr << cmppdsfile << " cannot be accessed\n"; exit(1); }

			// read compare_file pdsfile
			//
			read_pdsfile(cmppdsfile, cmpdir);

			// ensure compare_file is a panasync file
			//
			getbasename(compare_file, cmpbasename);
			if (cmpdir.find(cmpbasename) == cmpdir.end())
			{ cerr << "compare_file " << compare_file << " is not a panasync file\n"; exit(1); }
			
			// calculate md5 digest of compare_file
			//
			unsigned char md5[16];
			FILE* fp=fopen(compare_file, "r");
			md5_stream(fp, md5);
			fclose(fp);

			// check content modification of the compare file
			// and set it dirty
			//
			if (memcmp(md5, cmpdir[cmpbasename].md5, sizeof(md5)) != 0)
			{
				filedata cmpfdata(md5, cmpdir[cmpbasename].stamp);
				cmpdir.erase(cmpbasename);
				cmpfdata.stamp.set_dirty();
				cmpdir[cmpbasename]=cmpfdata;
			}


		}
		// ensure access to target pdsfile
		//
		if (access(pdsfile, R_OK) == -1)
		{ 
			// if (verbouse) cerr << "cannot access " << pdsfile << "\n"; i
			continue; 
		}

		if (!isfile) cout << getbasedir(pdsfile,tmpfilename) << ":" << endl;
		
		// read target pdsfile
		//
		panadir_t dir;
		read_pdsfile(pdsfile, dir);

		// start printing contents
		// 
		for (panadir_t::iterator i = dir.begin(); i != dir.end(); i++)
		{
			// If listing a file, only list that file
			//
			if (isfile && strcmp(i->first.c_str(),eachfile)!=0) continue;
			if (isfile) cout << getbasedir(pdsfile,tmpfilename) << ":" << endl;

			// Update md5 count, if possible
			// There is no need to write new md5s, since they must always be checked
			//
			unsigned char md5[16];
			getbasedir(pdsfile,tmpfilename);
			strcat(tmpfilename,i->first.c_str());
			FILE* fp=fopen(tmpfilename, "r");
			md5_stream(fp, md5);
			fclose(fp);

			if (memcmp(md5, dir[i->first].md5, sizeof(md5)) != 0)
			{
				dir[i->first].stamp.set_dirty();
				memcpy(dir[i->first].md5,md5,sizeof(md5));
			}
			
			
			if (compare_file)
			{
				cout << compare_file << " ";
				pts cmppts=cmpdir[cmpbasename].stamp;
				if (cmppts.uncomparable(i->second.stamp))
					cout << " ? ";
				else
					if (cmppts.equivalent(i->second.stamp))
						cout << "== ";
					else
						if (cmppts.dominates(i->second.stamp))
							cout << " > ";
						else
							if (i->second.stamp.dominates(cmppts))
								cout << " < ";
							else 
								cout << "|| ";
			}
			cout << i->first << " ";
			for (int j = 0; longopt && j < 16; ++j)
				printf ("%02x", i->second.md5[j]);
			if (longopt) cout << " ";
			cout << i->second.stamp;
			cout << endl;
		}

	}
	while (++optind < argc);
}

//	========================================================================
//	panamove
//	========================================================================
//

void	display_panamove_usage()
{
 	cerr <<
"usage: panamove {--help -h}\n"
"       panamove [--base_file -b] [file] [--force -f] source target\n";
}

void panamove(int argc, char ** argv)
{
	char *	pt_source	= 0;
	char *	pt_target	= 0;
	bool	pt_force	= false;
	char *	pt_base_path	= PTDEF_BASE_PATH;

	// Process command-line options
	//
	while (1)
	{
		int short_option;	// Short option character

		// Long options declaration & short options mappings
		//
		static struct option long_options[] =
		{
			{ "base_file", 1, NULL, 'b' },
			{ "force", 0, NULL, 'f' },
			{ "help", 0, NULL, 'h' },
			{ 0, 0, 0, 0 }
		};

		// Check end of option list
		//
		if ((short_option = getopt_long (argc, argv, "b:fh", 
			long_options, NULL)) == -1) break;

		// Process short option
		//
		switch (short_option)
		{
			case 'b':
				pt_base_path = strdup(optarg);
				break;
			case 'f':
				pt_force = true;
				cerr << "panamove: '--force' not ";
				cerr << "implemented. ";
				cerr << "[internal error]\n";
				exit(2);
				break;
			case 'h':
				display_panamove_usage();
				exit(0);
				break;
			case ':':
				cerr << "panamove: missing parameter on '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			case '?':
				cerr << "panamove: extraneous parameter '";
				cerr << argv[optind] << "'\n";
				exit(1);
				break;
			default:
				cerr << "panamove: unexpected option '";
				cerr << short_option << "' ";
				cerr << "[internal error]\n";
				exit(2);
		}
	}

	// Retrieve source and target files from command-line
	//
	if (optind + 2 == argc)
	{
		pt_source = argv[optind++];
		pt_target = argv[optind];
	}
	else
	{
		cerr << "panamove: illegal command line arguments.\n";
		exit(1);
	}

	// Source and target related path names
	//
	char srcbasename[MAXPATHLEN];
	char srcbasedir[MAXPATHLEN];
	char srcpdsfile[MAXPATHLEN];
	char trgbasename[MAXPATHLEN];
	char trgbasedir[MAXPATHLEN];
	char trgpdsfile[MAXPATHLEN];
	getbasename(pt_source, srcbasename);
	getbasedir(pt_source, srcbasedir);
	getbasename(pt_target, trgbasename);
	getbasedir(pt_target, trgbasedir);
	strcpy(srcpdsfile, srcbasedir);
	strcat(srcpdsfile, PTDEF_PDS_FILENAME);
	strcpy(trgpdsfile, trgbasedir);
	strcat(trgpdsfile, PTDEF_PDS_FILENAME);

	// Ensure read access to the source file
	//
	if (access(pt_source, R_OK) == -1)
	{
		cerr << "panamove: cannot read from " << pt_source << "\n";
		exit(1);
	}

	// Ensure read/write access to the source directory
	//
	if (access(srcbasedir, R_OK+W_OK) == -1)
	{
		cerr << "panamove: " << srcbasedir;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	// Ensure read/write access to the source pdsfile
	//
	if (access(srcpdsfile, R_OK+W_OK) == -1)
	{
		cerr << "panamove: " << srcpdsfile;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	// Read source pdsfile
	//
	panadir_t srcdir;
	read_pdsfile(srcpdsfile, srcdir);

	// Ensure source is a panasync file
	//
	if (srcdir.find(srcbasename) == srcdir.end())
	{
		cerr << "panamove: source " << pt_source;
		cerr << " is not a panasync file.\n";
		exit(1);
	}

	// Ensure read/write access to the target directory
	//
	if (access(trgbasedir, R_OK+W_OK) == -1)
	{
		cerr << "panamove: " << trgbasedir;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	// Check existance of target pdsfile
	// if it exists ensure read/write access
	//
	if (access(trgpdsfile, F_OK) == 0
		&& access(trgpdsfile, R_OK+W_OK) == -1)
	{
		cerr << "panamove: " << trgpdsfile;
		cerr << " cannot be accessed for read/write.\n";
		exit(1);
	}

	//streamlock slpdsfile(pdsfile);

	// Read target pdsfile if exists
	//
	panadir_t trgdir;
	if (access(trgpdsfile, F_OK) == 0)
		read_pdsfile(trgpdsfile, trgdir);

	// Ensure target not a panasync file
	//
	if (trgdir.find(trgbasename) != trgdir.end())
	{
		cerr << "panamove: " << pt_target;
		cerr << " is a panasync file. Cannot overwrite.\n";
		cerr << "Do a 'panajoin' before 'panamove'.\n";
		exit(1);
	}

	// Ensure source and target are different files
	//
	if (access(pt_target, F_OK) == 0
		&& getinode(pt_source) == getinode(pt_target))
	{
		cerr << "panamove: source " << pt_source;
		cerr << " and target " << pt_target;
		cerr << " are the same file.\n";
		exit(1);
	}
	
	// Copy content
	//
	int ifs = open(pt_source, O_RDONLY);
	int ofs = creat(pt_target, 0666), n;
	char buffer[PTBUFSIZE];
	while ((n = read(ifs, buffer, sizeof(buffer))) > 0)
		assert(write (ofs, buffer, n) == n);
	close (ifs);
	close (ofs);

	// Update pdsfiles, but first check if pdsfiles are shared
	//
	if (access(trgpdsfile, F_OK) == 0
		&& getinode(srcpdsfile) == getinode(trgpdsfile))
	{
		if (strcmp(srcbasename, trgbasename) == 0)
		{
			cerr << "panamove: source " << pt_source;
			cerr << " and target " << pt_target;
			cerr << " would be the same file.\n";
			exit(1);
		}
		srcdir[trgbasename] = srcdir[srcbasename];
		srcdir.erase(srcbasename);
		write_pdsfile(srcpdsfile, srcdir);
	}
	else
	{
		trgdir[trgbasename] = srcdir[srcbasename];
		srcdir.erase(srcbasename);
		write_pdsfile(srcpdsfile, srcdir);
		write_pdsfile(trgpdsfile, trgdir);
	}

	// Delete source file
	//
	unlink(pt_source);

	// Free resources
	//
	if (pt_base_path != PTDEF_BASE_PATH) free(pt_base_path);

}

//	========================================================================
//	not properly implemented
//	========================================================================
//

void	pana(int argc, char ** argv) {}
void	panasync(int argc, char ** argv) {}

//	========================================================================
//	main
//	========================================================================
//

int	main(int argc, char ** argv)
{
	int c;
	char argv0[MAXPATHLEN];

	if (!strcmp(getbasename(argv[0], argv0), "panainit"))
		panainit(argc, argv);
	else if (!strcmp(getbasename(argv[0], argv0), "pananew"))
		pananew(argc, argv);
	else if (!strcmp(getbasename(argv[0], argv0), "panadup"))
		panadup(argc, argv);
	else if (!strcmp(getbasename(argv[0], argv0), "panajoin"))
		panajoin(argc, argv);
	else if (!strcmp(getbasename(argv[0], argv0), "panasync"))
		panasync(argc, argv);
	else if (!strcmp(getbasename(argv[0], argv0), "panalist"))
		panalist(argc, argv);
	else if (!strcmp(getbasename(argv[0], argv0), "panamove"))
		panamove(argc, argv);
	else if (!strcmp(getbasename(argv[0], argv0), "pana"))
		pana(argc, argv);
	else
	{
		cerr << "warning: pana, my name is pana!\n";
		pana(argc, argv);
	}

	return 0;
}
